#include "precompile.h"

#include <string.h>

#include "ss_proto.pb.h"
#include "ss_msgid.pb.h"
#include "MSConn.h"
#include <a8/tcpclient.h>
#include <a8/udplog.h>
#include <a8/timer.h>
#include <a8/ioloop.h>
#include <a8/asynctcpclient.h>

#include "app.h"
#include "playermgr.h"
#include "jsondatamgr.h"

const int PACK_MAX = 1024 * 64 * 2;

void MSConn::Init(int instance_id, const std::string& remote_ip, int remote_port)
{
    if (remote_ip.empty()) {
        abort();
    }
    this->instance_id = instance_id;
    this->remote_ip = remote_ip;
    this->remote_port = remote_port;

    recv_bufflen_ = 0;
    last_pong_tick = a8::XGetTickCount();
    recv_buff_ = (char*) malloc(PACK_MAX * 2);
    tcp_client_ = a8::IoLoop::Instance()->CreateAsyncTcpClient();
    tcp_client_->remote_address = remote_ip;
    tcp_client_->remote_port = remote_port;
    tcp_client_->on_error = std::bind(&MSConn::on_error, this, std::placeholders::_1, std::placeholders::_2);
    tcp_client_->on_connect = std::bind(&MSConn::on_connect, this, std::placeholders::_1);
    tcp_client_->on_disconnect = std::bind(&MSConn::on_disconnect, this, std::placeholders::_1);
    tcp_client_->on_socketread = std::bind(&MSConn::on_socketread, this ,std::placeholders::_1, std::placeholders::_2, std::placeholders::_3);
    a8::Timer::Instance()->AddRepeatTimerAndAttach
        (1000 * 9 + a8::RandEx(500, 150),
         a8::XParams().SetSender(this),
         [] (const a8::XParams& param)
         {
             MSConn* conn = (MSConn*)param.sender.GetUserData();
             conn->CheckAlive();
             conn->SyncIMServerList();
         },
         &timer_attacher_.timer_list_);
    a8::Timer::Instance()->AddRepeatTimerAndAttach
        (800 + a8::RandEx(100, 300),
         a8::XParams().SetSender(this),
         [] (const a8::XParams& param)
         {
             MSConn* conn = (MSConn*)param.sender.GetUserData();
             conn->ReportServerInfo();
         },
         &timer_attacher_.timer_list_);
}

void MSConn::UnInit()
{
    timer_attacher_.ClearTimerList();
    tcp_client_->Close();
    a8::IoLoop::Instance()->DestoryAsyncTcpClient(tcp_client_);
    tcp_client_ = nullptr;
    recv_bufflen_ = 0;
    free(recv_buff_);
    recv_buff_ = nullptr;
}

void MSConn::Open()
{
    tcp_client_->Open();
}

void MSConn::Close()
{
    tcp_client_->Close();
}

bool MSConn::Connected()
{
    return tcp_client_->Connected();
}

void MSConn::on_error(a8::AsyncTcpClient* sender, int errorId)
{
    a8::UdpLog::Instance()->Error("master server errorid=%d remote_ip:%s remote_port:%d",
                                  {
                                   errorId,
                                   sender->remote_address,
                                   sender->remote_port
                                  });
}

void MSConn::on_connect(a8::AsyncTcpClient* sender)
{
    recv_bufflen_ = 0;
    a8::UdpLog::Instance()->Info("master server connected remote_ip:%s remote_port:%d",
                                 {
                                   sender->remote_address,
                                   sender->remote_port
                                 });
    App::Instance()->AddIMMsg(IM_MSConnConnect,
                              a8::XParams()
                              .SetSender(instance_id)
                              );
}

void MSConn::on_disconnect(a8::AsyncTcpClient* sender)
{
    recv_bufflen_ = 0;
    a8::UdpLog::Instance()->Info("master server %d disconnected after 10s later reconnect "
                                 "remote_ip:%s remote_port:%d",
                                 {
                                  instance_id,
                                  sender->remote_address,
                                  sender->remote_port
                                 });
    App::Instance()->AddIMMsg(IM_MSConnDisconnect,
                              a8::XParams()
                              .SetSender(instance_id)
                              );
}

void MSConn::on_socketread(a8::AsyncTcpClient* sender, char* buf, unsigned int len)
{
    #if 0
    ++PerfMonitor::Instance()->read_count;
    #endif
    if (recv_bufflen_ + len > 2 * PACK_MAX) {
        recv_bufflen_ = 0;
        a8::UdpLog::Instance()->Debug("recvied master server too long message", {});
        return;
    } else {
        memmove(&recv_buff_[recv_bufflen_], buf, len);
        recv_bufflen_ += len;
    }

    bool warning = false;
    unsigned int offset = 0;
    while (recv_bufflen_ - offset >= sizeof(f8::PackHead)) {
        f8::PackHead* p = (f8::PackHead*) &recv_buff_[offset];
        if (p->magic_code == f8::MAGIC_CODE) {
            if (recv_bufflen_ - offset < sizeof(f8::PackHead) + p->packlen) {
                break;
            }
            App::Instance()->AddSocketMsg(SF_MSConn,
                                          0,
                                          instance_id,
                                          p->msgid,
                                          p->seqid,
                                          &recv_buff_[offset + sizeof(f8::PackHead)],
                                          p->packlen);
            offset += sizeof(f8::PackHead) + p->packlen;
        } else {
            warning = true;
            offset++;
            continue;
        }
    }

    if (warning) {
        a8::UdpLog::Instance()->Debug("recvied bad package", {});
    }
    if (offset > 0 && offset < recv_bufflen_) {
        memmove(recv_buff_, recv_buff_ + offset, recv_bufflen_ - offset);
    }
    recv_bufflen_ -= offset;
    #if 1
    last_pong_tick = a8::XGetTickCount();
    #endif
}

void MSConn::CheckAlive()
{
    if (!Connected()) {
        Open();
    } else {
        if (a8::XGetTickCount() - last_pong_tick > 60 * 10 * 1000) {
            last_pong_tick = a8::XGetTickCount();
            Close();
            Open();
        } else {
            ss::SS_Ping msg;
            SendMsg(msg);
        }
    }
}

void MSConn::SyncIMServerList()
{
    if (Connected()) {
        ss::SS_IM_IMServerList msg;
        SendMsg(msg);
    }
}

void MSConn::ReportServerInfo()
{
    if (Connected()) {
        ss::SS_IM_ReportServerInfo msg;
        msg.set_instance_id(App::Instance()->instance_id);
        msg.set_online_num(PlayerMgr::Instance()->OnlineNum());
        msg.set_ip(JsonDataMgr::Instance()->GetConf()->At("ip")->AsXValue().GetString());
        msg.set_port(JsonDataMgr::Instance()->GetConf()->At("wsproxy_port")->AsXValue().GetInt());
        msg.set_servicing(App::Instance()->servicing);
        SendMsg(msg);
    }
}

void MSConn::OnConnectSync()
{
    CheckAlive();
    SyncIMServerList();
    ReportServerInfo();
}

void MSConn::OnDisconnectSync()
{

}
