#include "precompile.h"

#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <mutex>
#include <condition_variable>

#include <a8/redis.h>
#include <a8/timer.h>
#include <a8/uuid.h>
#include <a8/ioloop.h>

#include "app.h"
#include "IMListener.h"
#include "WSListener.h"
#include "jsondatamgr.h"
#include "handlermgr.h"
#include "ss_msgid.pb.h"
#include "ss_proto.pb.h"
#include "playermgr.h"
#include "groupmgr.h"
#include "dbengine.h"
#include "dbhelper.h"
#include "synchelper.h"
#include "perfmonitor.h"
#include "gamelog.h"
#include "asynctaskmgr.h"
#include "guildmgr.h"
#include "metamgr.h"
#include "chatmgr.h"

#include "MSConnMgr.h"
#include "IMConnMgr.h"

#include "cs_msgid.pb.h"

#include "framework/cpp/msgqueue.h"
#include "framework/cpp/tglog.h"
#include "framework/cpp/netmsghandler.h"
#include "framework/cpp/httpclientpool.h"

struct MsgNode
{
    SocketFrom_e sockfrom;
    int sockhandle;
    unsigned short msgid;
    unsigned int seqid;
    long ip_saddr;
    char* buf;
    int buflen;
    MsgNode* next;
};

struct IMMsgNode
{
    unsigned short msgid;
    a8::XParams params;
    IMMsgNode* next = nullptr;

};

const char* const PROJ_LOG_ROOT_FMT = "/data/logs/%s/logs";
const char* const PROJ_LOG_FILENAME_FMT = "log_$pid_%Y%m%d.log";

bool App::Init(int argc, char* argv[])
{
    signal(SIGPIPE, SIG_IGN);
    this->argc = argc;
    this->argv = argv;

    if (!ParseOpt()) {
        terminated = true;
        if (instance_id <= 0) {
            a8::XPrintf("friend_imserver启动失败,缺少-i参数\n", {});
        } else if (instance_id > MAX_INSTANCE_ID) {
            a8::XPrintf("friend_imserver启动失败,-i参数不能大于%d\n", {MAX_INSTANCE_ID});
        }
        return false;
    }
    a8::XPrintf("friend_imserver starting instance_id:%d pid:%d\n",
                {
                 instance_id,
                 getpid(),
                });

    loop_mutex_ = new std::mutex();
    loop_cond_ = new std::condition_variable();
    msg_mutex_ = new std::mutex();
    im_msg_mutex_ = new std::mutex();

    srand(time(nullptr));
    InitLog();
    HandlerMgr::Instance()->Init();
    a8::Timer::Instance()->Init();
    PerfMonitor::Instance()->Init();
    f8::MsgQueue::Instance()->Init();
    f8::TGLog::Instance()->Init(a8::Format(PROJ_NAME_FMT, {GAME_ID}), true);
    f8::HttpClientPool::Instance()->Init(MAX_ALL_HTTP_NUM, MAX_SYS_HTTP_NUM, MAX_USER_HTTP_NUM);
    GameLog::Instance()->Init();
    a8::IoLoop::Instance()->Init(1);
    JsonDataMgr::Instance()->Init();
    MetaMgr::Instance()->Init();
    uuid.SetMachineId(instance_id);
    IMConnMgr::Instance()->Init();
    MSConnMgr::Instance()->Init();
    WSListener::Instance()->Init();
    IMListener::Instance()->Init();
    PlayerMgr::Instance()->Init();
    GroupMgr::Instance()->Init();
    DBEngine::Instance()->Init();
    DBHelper::Instance()->Init();
    SyncHelper::Instance()->Init();
    AsyncTaskMgr::Instance()->Init();
    GuildMgr::Instance()->Init();
    ChatMgr::Instance()->Init();

    a8::UdpLog::Instance()->Info("friend_imserver starting instance_id:%d pid:%d 13423",
                                 {
                                  instance_id,
                                  getpid(),
                                 });
    if (HasFlag(1)) {
        a8::Timer::Instance()->AddDeadLineTimer(
                                                1000 * 60,
                                                a8::XParams(),
                                                [] (const a8::XParams& param)
                                                {
                                                    App::Instance()->terminated = true;
                                                    App::Instance()->NotifyLoopCond();
                                                }
                                                );
    }
    if (HasFlag(4)) {
        a8::Timer::Instance()->AddDeadLineTimer(
                                                1000 * 30,
                                                a8::XParams(),
                                                [] (const a8::XParams& param)
                                                {
                                                    App::Instance()->shutdowned = true;
                                                    a8::XPrintf("shutdowned\n", {});
                                                }
                                                );
    }
    return true;
}

void App::UnInit()
{
    a8::XPrintf("friend_imserver terminating instance_id:%d pid:%d\n", {instance_id, getpid()});
    ChatMgr::Instance()->UnInit();
    GuildMgr::Instance()->UnInit();
    AsyncTaskMgr::Instance()->UnInit();
    SyncHelper::Instance()->UnInit();
    DBHelper::Instance()->UnInit();
    DBEngine::Instance()->UnInit();
    GroupMgr::Instance()->UnInit();
    PlayerMgr::Instance()->UnInit();
    IMListener::Instance()->UnInit();
    WSListener::Instance()->UnInit();
    IMConnMgr::Instance()->UnInit();
    MSConnMgr::Instance()->UnInit();
    MetaMgr::Instance()->UnInit();
    JsonDataMgr::Instance()->UnInit();
    f8::HttpClientPool::Instance()->UnInit();
    a8::IoLoop::Instance()->UnInit();
    GameLog::Instance()->UnInit();
    f8::TGLog::Instance()->UnInit();
    f8::MsgQueue::Instance()->UnInit();
    PerfMonitor::Instance()->UnInit();
    a8::Timer::Instance()->UnInit();
    HandlerMgr::Instance()->UnInit();
    UnInitLog();

    FreeSocketMsgQueue();
    FreeIMMsgQueue();
    A8_SAFE_DELETE(im_msg_mutex_);
    A8_SAFE_DELETE(msg_mutex_);
    A8_SAFE_DELETE(loop_cond_);
    A8_SAFE_DELETE(loop_mutex_);
    a8::XPrintf("friend_imserver terminated instance_id:%d pid:%d\n", {instance_id, getpid()});
}

int App::Run()
{
    int ret = 0;
    a8::UdpLog::Instance()->Info("friend_imserver running", {});
    while (!terminated) {
        a8::tick_t begin_tick = a8::XGetTickCount();
        nowtime = time(nullptr);
        QuickExecute();
        SlowerExecute();
        a8::tick_t end_tick = a8::XGetTickCount();
        if (end_tick - begin_tick > PerfMonitor::Instance()->max_run_delay_time) {
            PerfMonitor::Instance()->max_run_delay_time = end_tick - begin_tick;
        }
        Schedule();
    }
    return ret;
}

void App::AddSocketMsg(SocketFrom_e sockfrom,
                       int sockhandle,
                       long ip_saddr,
                       unsigned short msgid,
                       unsigned int seqid,
                       const char *msgbody,
                       int bodylen)
{
    MsgNode *p = (MsgNode*) malloc(sizeof(MsgNode));
    memset(p, 0, sizeof(MsgNode));
    p->sockfrom = sockfrom;
    p->ip_saddr = ip_saddr;
    p->sockhandle = sockhandle;
    p->msgid = msgid;
    p->seqid = seqid;
    p->buf = nullptr;
    p->buflen = bodylen;
    if (bodylen > 0) {
        p->buf = (char*)malloc(bodylen);
        memmove(p->buf, msgbody, bodylen);
    }
    msg_mutex_->lock();
    if (bot_node_) {
        bot_node_->next = p;
        bot_node_ = p;
    } else {
        top_node_ = p;
        bot_node_ = p;
    }
    ++msgnode_size_;
    msg_mutex_->unlock();
    NotifyLoopCond();
}

void App::AddIMMsg(unsigned short imcmd, a8::XParams params)
{
    IMMsgNode *p = new IMMsgNode;
    p->msgid = imcmd;
    p->params = params;
    p->next = nullptr;
    im_msg_mutex_->lock();
    if (im_bot_node_) {
        im_bot_node_->next = p;
        im_bot_node_ = p;
    } else {
        im_top_node_ = p;
        im_bot_node_ = p;
    }
    im_msg_mutex_->unlock();
    NotifyLoopCond();
}

void App::QuickExecute()
{
    ProcessIMMsg();
    DispatchMsg();
    PlayerMgr::Instance()->Update(a8::XGetTickCount());
    a8::Timer::Instance()->Update();
}

void App::SlowerExecute()
{
}

void App::NotifyLoopCond()
{
    std::unique_lock<std::mutex> lk(*loop_mutex_);
    loop_cond_->notify_all();
}

void App::Schedule()
{
    std::unique_lock<std::mutex> lk(*loop_mutex_);
    if (!HasTask()) {
        int sleep_time = a8::Timer::Instance()->GetIdleableMillSeconds();
        loop_cond_->wait_for(lk, std::chrono::milliseconds(sleep_time));
        if (sleep_time > PerfMonitor::Instance()->max_timer_idle) {
            PerfMonitor::Instance()->max_timer_idle = sleep_time;
        }
    }
}

bool App::HasTask()
{
    {
        if (!im_work_node_) {
            im_msg_mutex_->lock();
            if (!im_work_node_ && im_top_node_) {
                im_work_node_ = im_top_node_;
                im_top_node_ = nullptr;
                im_bot_node_ = nullptr;
            }
            im_msg_mutex_->unlock();
        }
        if (im_work_node_) {
            return true;
        }
    }
    {
        if (!work_node_) {
            msg_mutex_->lock();
            if (!work_node_ && top_node_) {
                work_node_ = top_node_;
                top_node_ = nullptr;
                bot_node_ = nullptr;
            }
            msg_mutex_->unlock();
        }
        if (work_node_) {
            return true;
        }
    }
    return false;
}

void App::DispatchMsg()
{
    long long starttick = a8::XGetTickCount();
    if (!work_node_ && top_node_) {
        msg_mutex_->lock();
        work_node_ = top_node_;
        top_node_ = nullptr;
        bot_node_ = nullptr;
        working_msgnode_size_ = msgnode_size_;
        msg_mutex_->unlock();
    }

    f8::MsgHdr hdr;
    while (work_node_) {
        MsgNode *pdelnode = work_node_;
        work_node_ = pdelnode->next;
        hdr.msgid = pdelnode->msgid;
        hdr.seqid = pdelnode->seqid;
        hdr.socket_handle = pdelnode->sockhandle;
        hdr.buf = pdelnode->buf;
        hdr.buflen = pdelnode->buflen;
        hdr.offset = 0;
        hdr.ip_saddr = pdelnode->ip_saddr;
        switch (pdelnode->sockfrom) {
        case SF_WSProxy:
            {
                ProcessWSProxyMsg(hdr);
            }
            break;
        case SF_IMServer:
            {
                ProcessIMServerMsg(hdr);
            }
            break;
        case SF_IMConn:
            {
                ProcessIMConnMsg(hdr);
            }
            break;
        case SF_MSConn:
            {
                ProcessMSConnMsg(hdr);
            }
            break;
        default:
            {

            }
            break;
        }
        if (pdelnode->buf) {
            free(pdelnode->buf);
        }
        free(pdelnode);
        working_msgnode_size_--;
        if (a8::XGetTickCount() - starttick > 200) {
            break;
        }
    }//end while

    if (!work_node_) {
        working_msgnode_size_ = 0;
    }
}

void App::ProcessWSProxyMsg(f8::MsgHdr& hdr)
{
    if (hdr.msgid == cs::CMMessageId_e::_CMLogin) {
        cs::CMLoginCommonHead common_head;
        bool ok = common_head.ParseFromArray(hdr.buf + hdr.offset, hdr.buflen - hdr.offset);
        if (ok) {
            if (!common_head.has_server_id()) {
                cs::CMLoginOld msg;
                msg.ParseFromArray(hdr.buf + hdr.offset, hdr.buflen - hdr.offset);
                PlayerMgr::Instance()->_CMLoginOld(hdr, msg);
            } else {
                cs::CMLogin msg;
                msg.ParseFromArray(hdr.buf + hdr.offset, hdr.buflen - hdr.offset);
                PlayerMgr::Instance()->_CMLogin(hdr, msg);
            }
        }
        return;
    }
    f8::NetMsgHandler* handler = f8::GetNetMsgHandler(&HandlerMgr::Instance()->wsmsghandler,
                                                      hdr.msgid);
    if (handler) {
        switch (handler->handlerid) {
        case HID_WSListener:
            {
                ProcessNetMsg(handler, WSListener::Instance(), hdr);
            }
            break;
        case HID_Player:
            {
                Player* hum = PlayerMgr::Instance()->GetPlayerBySocket(hdr.socket_handle);
                if (hum) {
                    ProcessNetMsg(handler, hum, hdr);
                }
            }
            break;
        case HID_PlayerMgr:
            {
                ProcessNetMsg(handler, PlayerMgr::Instance(), hdr);
            }
            break;
        default:
            {
            }
            break;
        }
        #ifdef DEBUG
        f8::DumpMsgToLog(hdr, handler, ">>>>>>WSP ");
        #endif
    }
}

void App::ProcessIMServerMsg(f8::MsgHdr& hdr)
{
    f8::NetMsgHandler* handler = f8::GetNetMsgHandler(&HandlerMgr::Instance()->immsghandler,
                                                      hdr.msgid);
    if (handler) {
        switch (handler->handlerid) {
        case HID_GuildMgr:
            ProcessNetMsg(handler, GuildMgr::Instance(), hdr);
            break;
        case HID_IMListener:
            ProcessNetMsg(handler, IMListener::Instance(), hdr);
            break;
        case HID_SyncHelper:
            ProcessNetMsg(handler, SyncHelper::Instance(), hdr);
            break;
        case HID_PlayerMgr:
            ProcessNetMsg(handler, PlayerMgr::Instance(), hdr);
            break;
        default:
            {
            }
            break;
        }
        #ifdef DEBUG
        f8::DumpMsgToLog(hdr, handler, ">>>>>>IMS ");
        #endif
    }
}

void App::ProcessIMConnMsg(f8::MsgHdr& hdr)
{
    f8::NetMsgHandler* handler = f8::GetNetMsgHandler(&HandlerMgr::Instance()->imcmsghandler,
                                                      hdr.msgid);
    if (handler) {
        switch (handler->handlerid) {
        case HID_GuildMgr:
            ProcessNetMsg(handler, GuildMgr::Instance(), hdr);
            break;
        case HID_PlayerMgr:
            ProcessNetMsg(handler, PlayerMgr::Instance(), hdr);
            break;
        default:
            {
            }
            break;
        }
        #ifdef DEBUG
        f8::DumpMsgToLog(hdr, handler, ">>>>>>IMC ");
        #endif
    }
}

void App::ProcessMSConnMsg(f8::MsgHdr& hdr)
{
    f8::NetMsgHandler* handler = f8::GetNetMsgHandler(&HandlerMgr::Instance()->mscmsghandler,
                                                      hdr.msgid);
    if (handler) {
        switch (handler->handlerid) {
        case HID_PlayerMgr:
            ProcessNetMsg(handler, PlayerMgr::Instance(), hdr);
            break;
        case HID_GroupMgr:
            ProcessNetMsg(handler, GroupMgr::Instance(), hdr);
            break;
        default:
            {
            }
            break;
        }
        #ifdef DEBUG
        f8::DumpMsgToLog(hdr, handler, ">>>>>>MSC ");
        #endif
    }
}

void App::ProcessIMMsg()
{
    if (!im_work_node_ && im_top_node_) {
        im_msg_mutex_->lock();
        im_work_node_ = im_top_node_;
        im_top_node_ = nullptr;
        im_bot_node_ = nullptr;
        im_msg_mutex_->unlock();
    }
    while (im_work_node_) {
        IMMsgNode *pdelnode = im_work_node_;
        switch (im_work_node_->msgid) {
        case f8::IM_SysMsgQueue:
            {
                const a8::XParams* param = (const a8::XParams*)pdelnode->params.param1.GetUserData();
                f8::MsgQueue::Instance()->ProcessMsg(pdelnode->params.sender.GetInt(),
                                                     *param
                                                     );
                delete param;
            }
            break;
        case IM_WSProxyDisconnect:
            {
                PlayerMgr::Instance()->OnWSProxyDisconnect(pdelnode->params);
            }
            break;
        case IM_MSConnConnect:
            {
                MSConnMgr::Instance()->__IM_MSConnConnect(pdelnode->params);
            }
            break;
        case IM_MSConnDisconnect:
            {
                MSConnMgr::Instance()->__IM_MSConnDisconnect(pdelnode->params);
            }
            break;
            #if 0
        case IM_TargetConnConnect:
            {
                GameClientMgr::Instance()->OnTargetServerConnect(pdelnode->params);
                TargetConn* conn = TargetConnMgr::Instance()->GetConnById(pdelnode->params.sender);
                if (conn && conn->Connected()) {
                    conn->SendStockMsg();
                }
            }
            break;
        case IM_TargetConnDisconnect:
            {
                GameClientMgr::Instance()->OnTargetServerDisconnect(pdelnode->params);
            }
            break;
            #endif
        case IM_ExecGM:
            {
                HandlerMgr::Instance()->ProcGMMsg(pdelnode->params.param3,
                                                  pdelnode->params.sender,
                                                  pdelnode->params.param1.GetString(),
                                                  pdelnode->params.param2.GetString()
                                                  );
            }
            break;
        }
        im_work_node_ = im_work_node_->next;
        delete pdelnode;
    }
}

void App::InitLog()
{
    std::string filename_fmt = PROJ_LOG_FILENAME_FMT;
    a8::ReplaceString(filename_fmt, "$pid", a8::XValue(getpid()));

    std::string proj_root_dir = a8::Format(PROJ_ROOT_FMT, {a8::Format(PROJ_NAME_FMT,{})});
    std::string proj_log_root_dir = a8::Format(PROJ_LOG_ROOT_FMT, {a8::Format(PROJ_NAME_FMT, {})});
    std::string log_file_name = a8::Format(PROJ_LOG_ROOT_FMT,
                                           {a8::Format(PROJ_NAME_FMT, {})}) + "/" + filename_fmt;

    a8::MkDir(proj_root_dir);
    a8::MkDir(proj_log_root_dir);
    a8::UdpLog::Instance()->SetLogFileName(log_file_name);
    a8::UdpLog::Instance()->Init();
    a8::UdpLog::Instance()->Info("proj_root_dir:%s", {proj_root_dir});
    a8::UdpLog::Instance()->Info("proj_log_root_dir:%s", {proj_log_root_dir});
    a8::UdpLog::Instance()->Info("log_file_name:%s", {log_file_name});
}

void App::UnInitLog()
{
    a8::UdpLog::Instance()->UnInit();
}

bool App::ParseOpt()
{
    int ch = 0;
    while ((ch = getopt(argc, argv, "n:i:f:")) != -1) {
        switch (ch) {
        case 'i':
            {
                instance_id = a8::XValue(optarg);
            }
            break;
        case 'f':
            {
                std::vector<std::string> strings;
                a8::Split(optarg, strings, ',');
                for (auto& str : strings) {
                    flags.insert(a8::XValue(str).GetInt());
                }
            }
            break;
        }
    }
    return instance_id > 0;
}

a8::XParams* App::AddContext(long long context_id)
{
    context_hash_[context_id] = a8::XParams();
    return GetContext(context_id);
}

void App::DelContext(long long context_id)
{
    context_hash_.erase(context_id);
}

a8::XParams* App::GetContext(long long context_id)
{
    auto itr = context_hash_.find(context_id);
    return itr != context_hash_.end() ? &(itr->second) : nullptr;
}

bool App::HasFlag(int flag)
{
    return flags.find(flag) != flags.end();
}

void App::FreeSocketMsgQueue()
{
    msg_mutex_->lock();
    if (!work_node_) {
        work_node_ = top_node_;
        top_node_ = nullptr;
        bot_node_ = nullptr;
    }
    while (work_node_) {
        MsgNode* pdelnode = work_node_;
        work_node_ = work_node_->next;
        if (pdelnode->buf) {
            free(pdelnode->buf);
        }
        free(pdelnode);
        if (!work_node_) {
            work_node_ = top_node_;
            top_node_ = nullptr;
            bot_node_ = nullptr;
        }
    }
    msg_mutex_->unlock();
}

void App::FreeIMMsgQueue()
{
    im_msg_mutex_->lock();
    if (!im_work_node_) {
        im_work_node_ = im_top_node_;
        im_top_node_ = nullptr;
        im_bot_node_ = nullptr;
    }
    while (im_work_node_) {
        IMMsgNode* pdelnode = im_work_node_;
        im_work_node_ = im_work_node_->next;
        if (pdelnode->msgid == f8::IM_SysMsgQueue) {
            a8::XParams* param = (a8::XParams*)pdelnode->params.param1.GetUserData();
            delete param;
        }
        delete pdelnode;
        if (!im_work_node_) {
            im_work_node_ = im_top_node_;
            im_top_node_ = nullptr;
            im_bot_node_ = nullptr;
        }
    }
    im_msg_mutex_->unlock();
}

long long App::NewUUID()
{
    return uuid.Generate();
}

long long App::NewSeqId()
{
    return ++seq_id_;
}

time_t App::BetweenDays(time_t time1, time_t time2)
{
    const int g_time_zone = 8;
    return (time1 + g_time_zone*3600)/3600/24 - (time2 + g_time_zone*3600)/3600/24;
}

bool App::IsTimeToReset(int time)
{
    return BetweenDays(nowtime - 60 * SYS_RESET_TIME, time - 60 * SYS_RESET_TIME) > 0;
}

void App::PreProcAvatarUrl(int self_channel, int target_channel, std::string& target_avatar_url)
{
    if (self_channel == 6001 && target_channel != 6001) {
        target_avatar_url = "https://wx.qlogo.cn/mmopen/vi_32/q4oRsMFYBwPEVAeVI7tiasWSqaibr0GPQia432JhibGRYhqqEJofpWDYxJPq6q0hQ0j7icdACdHL78hrjYYHSjZQ3YA/132";
    }
}
